### Source:  http://www.ucolick.org/~xavier/DLA/Kingpin/atomic.dat ####
linelist_highz = {
    'H I 1216'    : {'pltLabel':'H-I'       ,"lambda": 1215.6701, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.41640  },
    'Si II 1260'  : {'pltLabel':'SI-II'     ,"lambda": 1260.4221, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.0070   },
    'O I 1302'    : {'pltLabel':'O-I+Si-II' ,"lambda": 1302.1685, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.04887  },
    'Si II 1304'  : {'pltLabel':''          ,"lambda": 1304.3702, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0940   },
    'C II  1334'  : {'pltLabel':'C-II'      ,"lambda": 1334.5323, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1278   },
    'Si II 1526'  : {'pltLabel':'Si-II'     ,"lambda": 1526.7066, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1270   },
    'Fe II 1608'  : {'pltLabel':'Fe-II'     ,"lambda": 1608.4511, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0580   },
    'Fe II 1611'  : {'pltLabel':''          ,"lambda": 1611.2005, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.00136  },

    'Al II 1670'  : {'pltLabel':'Al-II'     ,"lambda": 1670.7874, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.8800   },
    'Fe II 2344'  : {'pltLabel':'Fe-II'     ,"lambda": 2344.214,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1140   },
    'Fe II 2374'  : {'pltLabel':'Fe-II'     ,"lambda": 2374.4612, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0313   },
    'Fe II 2382'  : {'pltLabel':'Fe-II'     ,"lambda": 2382.765,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.320    },
    'Fe II 2586'  : {'pltLabel':'Fe-II'     ,"lambda": 2586.6500, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0691   },
    'Fe II 2600'  : {'pltLabel':'Fe-II'     ,"lambda": 2600.1729, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.2390   },
    'Mg II 2796'  : {'pltLabel':'Mg-II'     ,"lambda": 2796.352,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.6123   },
    'Mg II 2803'  : {'pltLabel':''          ,"lambda": 2803.531,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.3054   },
    'Mg I 2852'   : {'pltLabel':'Mg-I'      ,"lambda": 2852.9642, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.8100   },

    'Ni II 1317'  : {'pltLabel':'Ni-II'     ,"lambda": 1317.217 , "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.07786  },
    'Ni II 1370'  : {'pltLabel':'Ni-II'     ,"lambda": 1370.131,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0769   },
    'Ni II 1709'  : {'pltLabel':'Ni-II'     ,"lambda": 1709.6042, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0324   },
    'Ni II 1741'  : {'pltLabel':'Ni-II'     ,"lambda": 1741.5531, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0427   },
    'Ni II 1751'  : {'pltLabel':''          ,"lambda": 1751.915,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0277   },
    'Si I 1845'   : {'pltLabel':'Si-I'      ,"lambda": 1845.520,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.2290   },
    'S II 1250'   : {'pltLabel':'S-II'      ,"lambda": 1250.584,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.005453 },
    'S II 1253'   : {'pltLabel':'S-II'      ,"lambda": 1253.811,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.01088  },
    'Si II 1808'  : {'pltLabel':'Si-II'     ,"lambda": 1808.0130, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.002186  },

    'Si IV 1393'    : {'pltLabel': 'Si-IV'    ,"lambda": 1393.755 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.5280  },
    'Si IV 1402'    : {'pltLabel': 'SI-IV'    ,"lambda": 1402.770 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.2620  },
    'C IV  1548'    : {'pltLabel': 'C-IV'     ,"lambda": 1548.195 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.19080 },
    'C IV  1550'    : {'pltLabel': ''         ,"lambda": 1550.770 , "highion": True, "plot": True, "emission": False, "comments":'complicated as there is ISM absorption, and broad stellar P-Cygni absorption + emission', "fion": 0.095220},
    'Al III 1854'   : {'pltLabel': 'Al-III'   ,"lambda": 1854.7164, "highion": True, "plot": True, "emission": False, "comments":'', "fion":0.5390   },
    'Al III 1862'   : {'pltLabel': ''         ,"lambda": 1862.7895, "highion": True, "plot": True, "emission": False, "comments":'', "fion":0.2680   },
}
